"""
nanopy.rpc
##########
A wrapper to make RPC requests to a node.
"""

from abc import ABC, abstractmethod
from typing import Any
import json

try:
    import requests

    ENABLE_HTTP = True
except ModuleNotFoundError:
    ENABLE_HTTP = False

try:
    import websocket

    ENABLE_WS = True
except ModuleNotFoundError:
    ENABLE_WS = False


class RPC(ABC):
    "RPC base class"

    @abstractmethod
    def request(self, data: dict[str, Any]) -> Any:
        """Make RPC request to nano node.
        Overridden in derived class

        :arg data: dict like object
        :return: JSON reponse as dict
        """
        raise NotImplementedError("Implement in a derived class")

    def account_balance(self, account: str, include_only_confirmed: bool = True) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_balance"
        data: dict[str, Any] = {}
        data["action"] = "account_balance"
        data["account"] = account
        if not include_only_confirmed:
            data["include_only_confirmed"] = False
        return self.request(data)

    def account_block_count(self, account: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_block_count"
        data: dict[str, Any] = {}
        data["action"] = "account_block_count"
        data["account"] = account
        return self.request(data)

    def accountget(self, key: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#accountget"
        data: dict[str, Any] = {}
        data["action"] = "accountget"
        data["key"] = key
        return self.request(data)

    def account_history(
        self,
        account: str,
        count: int = 1,
        raw: bool = False,
        head: str = "",
        include_linked_account: bool = False,
        offset: int = 0,
        reverse: bool = False,
        account_filter: list[str] | None = None,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_history"
        data: dict[str, Any] = {}
        data["action"] = "account_history"
        data["account"] = account
        data["count"] = count
        if raw:
            data["raw"] = True
        if head:
            data["head"] = head
        if include_linked_account:
            data["include_linked_account"] = True
        if offset:
            data["offset"] = offset
        if reverse:
            data["reverse"] = reverse
        if account_filter:
            data["account_filter"] = account_filter
        return self.request(data)

    def account_info(
        self,
        account: str,
        include_confirmed: bool = False,
        representative: bool = False,
        weight: bool = False,
        pending: bool = False,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_info"
        data: dict[str, Any] = {}
        data["action"] = "account_info"
        data["account"] = account
        if include_confirmed:
            data["include_confirmed"] = True
        if representative:
            data["representative"] = True
        if weight:
            data["weight"] = True
        if pending:
            data["pending"] = True
        return self.request(data)

    def account_key(self, account: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_key"
        data: dict[str, Any] = {}
        data["action"] = "account_key"
        data["account"] = account
        return self.request(data)

    def account_representative(self, account: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_representative"
        data: dict[str, Any] = {}
        data["action"] = "account_representative"
        data["account"] = account
        return self.request(data)

    def account_weight(self, account: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_weight"
        data: dict[str, Any] = {}
        data["action"] = "account_weight"
        data["account"] = account
        return self.request(data)

    def accounts_balances(
        self, accounts: list[str], include_only_confirmed: bool = True
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#accounts_balances"
        data: dict[str, Any] = {}
        data["action"] = "accounts_balances"
        data["accounts"] = accounts
        if not include_only_confirmed:
            data["include_only_confirmed"] = False
        return self.request(data)

    def accounts_frontiers(self, accounts: list[str]) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#accounts_frontiers"
        data: dict[str, Any] = {}
        data["action"] = "accounts_frontiers"
        data["accounts"] = accounts
        return self.request(data)

    def accounts_receivable(
        self,
        accounts: list[str],
        count: int = 1,
        threshold: str = "",
        source: bool = False,
        include_active: bool = False,
        sorting: bool = False,
        include_only_confirmed: bool = True,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#accounts_receivable"
        data: dict[str, Any] = {}
        data["action"] = "accounts_receivable"
        data["accounts"] = accounts
        data["count"] = count
        if threshold:
            data["threshold"] = threshold
        if source:
            data["source"] = True
        if include_active:
            data["include_active"] = True
        if sorting:
            data["sorting"] = True
        if not include_only_confirmed:
            data["include_only_confirmed"] = False
        return self.request(data)

    def accounts_representatives(self, accounts: list[str]) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#accounts_representatives"
        data: dict[str, Any] = {}
        data["action"] = "accounts_representatives"
        data["accounts"] = accounts
        return self.request(data)

    def available_supply(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#available_supply"
        data: dict[str, Any] = {}
        data["action"] = "available_supply"
        return self.request(data)

    def block_account(self, _hash: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#block_account"
        data: dict[str, Any] = {}
        data["action"] = "block_account"
        data["hash"] = _hash
        return self.request(data)

    def block_confirm(self, _hash: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#block_confirm"
        data: dict[str, Any] = {}
        data["action"] = "block_confirm"
        data["hash"] = _hash
        return self.request(data)

    def block_count(self, include_cemented: bool = True) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#block_count"
        data: dict[str, Any] = {}
        data["action"] = "block_count"
        if not include_cemented:
            data["include_cemented"] = False
        return self.request(data)

    def block_create(
        self,
        balance: str,
        representative: str,
        previous: str,
        wallet: str = "",
        account: str = "",
        key: str = "",
        source: str = "",
        destination: str = "",
        link: str = "",
        work: str = "",
        version: str = "work_1",
        json_block: bool = False,
        difficulty: str = "",
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#block_create"
        data: dict[str, Any] = {}
        data["action"] = "block_create"
        data["type"] = "state"
        data["balance"] = balance
        if wallet:
            data["wallet"] = wallet
        if account:
            data["account"] = account
        if key:
            data["key"] = key
        if source:
            data["source"] = source
        if destination:
            data["destination"] = destination
        if link:
            data["link"] = link
        data["representative"] = representative
        data["previous"] = previous
        if work:
            data["work"] = work
        elif difficulty:
            data["difficulty"] = difficulty
        if version in ["work_1"]:
            data["version"] = version
        if json_block:
            data["json_block"] = True
        return self.request(data)

    def block_hash(self, block: dict[str, str], json_block: bool = False) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#block_hash"
        data: dict[str, Any] = {}
        data["action"] = "block_hash"
        data["block"] = block
        if json_block:
            data["json_block"] = True
        return self.request(data)

    def block_info(
        self, _hash: str, json_block: bool = False, include_linked_account: bool = False
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#block_info"
        data: dict[str, Any] = {}
        data["action"] = "block_info"
        data["hash"] = _hash
        if json_block:
            data["json_block"] = True
        if include_linked_account:
            data["include_linked_account"] = True
        return self.request(data)

    def blocks(self, hashes: list[str], json_block: bool = False) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#blocks"
        data: dict[str, Any] = {}
        data["action"] = "blocks"
        data["hashes"] = hashes
        if json_block:
            data["json_block"] = True
        return self.request(data)

    def blocks_info(
        self,
        hashes: list[str],
        include_linked_account: bool = False,
        pending: bool = False,
        source: bool = False,
        receive_hash: bool = False,
        json_block: bool = False,
        include_not_found: bool = False,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#blocks_info"
        data: dict[str, Any] = {}
        data["action"] = "blocks_info"
        data["hashes"] = hashes
        if include_linked_account:
            data["include_linked_account"] = True
        if pending:
            data["pending"] = True
        if source:
            data["source"] = True
        if receive_hash:
            data["receive_hash"] = True
        if json_block:
            data["json_block"] = True
        if include_not_found:
            data["include_not_found"] = True
        return self.request(data)

    def bootstrap(
        self,
        address: str,
        port: str,
        bypass_frontier_confirmation: bool = False,
        _id: str = "",
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#bootstrap"
        data: dict[str, Any] = {}
        data["action"] = "bootstrap"
        data["address"] = address
        data["port"] = port
        if _id:
            data["id"] = _id
        if bypass_frontier_confirmation:
            data["bypass_frontier_confirmation"] = True
        return self.request(data)

    def bootstrap_any(
        self, force: bool = False, _id: str = "", account: str = ""
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#bootstrap_any"
        data: dict[str, Any] = {}
        data["action"] = "bootstrap_any"
        if force:
            data["force"] = True
        if _id:
            data["id"] = _id
        if account:
            data["account"] = account
        return self.request(data)

    def bootstrap_lazy(self, hash_: str, force: bool = False, _id: str = "") -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#bootstrap_lazy"
        data: dict[str, Any] = {}
        data["action"] = "bootstrap_lazy"
        data["hash"] = hash_
        if force:
            data["force"] = True
        if _id:
            data["id"] = _id
        return self.request(data)

    def bootstrap_priorities(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#bootstrap_priorities"
        data: dict[str, Any] = {}
        data["action"] = "bootstrap_priorities"
        return self.request(data)

    def bootstrap_reset(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#bootstrap_reset"
        data: dict[str, Any] = {}
        data["action"] = "bootstrap_reset"
        return self.request(data)

    def bootstrap_status(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#bootstrap_status"
        data: dict[str, Any] = {}
        data["action"] = "bootstrap_status"
        return self.request(data)

    def chain(
        self, block: str, count: int = 1, offset: int = 0, reverse: bool = False
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#chain"
        data: dict[str, Any] = {}
        data["action"] = "chain"
        data["block"] = block
        data["count"] = count
        if offset:
            data["offset"] = offset
        if reverse:
            data["reverse"] = True
        return self.request(data)

    def confirmation_active(self, announcements: int = 0) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#confirmation_active"
        data: dict[str, Any] = {}
        data["action"] = "confirmation_active"
        if announcements:
            data["announcements"] = announcements
        return self.request(data)

    def confirmation_height_currently_processing(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#confirmation_height_currently_processing"
        data: dict[str, Any] = {}
        data["action"] = "confirmation_height_currently_processing"
        return self.request(data)

    def confirmation_history(self, _hash: str = "") -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#confirmation_history"
        data: dict[str, Any] = {}
        data["action"] = "confirmation_history"
        if _hash:
            data["hash"] = _hash
        return self.request(data)

    def confirmation_info(
        self,
        root: str,
        contents: bool = True,
        representatives: bool = False,
        json_block: bool = False,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#confirmation_info"
        data: dict[str, Any] = {}
        data["action"] = "confirmation_info"
        data["root"] = root
        if not contents:
            data["contents"] = False
        if representatives:
            data["representatives"] = True
        if json_block:
            data["json_block"] = True
        return self.request(data)

    def confirmation_quorum(self, peer_details: bool = False) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#confirmation_quorum"
        data: dict[str, Any] = {}
        data["action"] = "confirmation_quorum"
        if peer_details:
            data["peer_details"] = True
        return self.request(data)

    def database_txn_tracker(self, min_read_time: int, min_write_time: int) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#database_txn_tracker"
        data: dict[str, Any] = {}
        data["action"] = "database_txn_tracker"
        data["min_read_time"] = min_read_time
        data["min_write_time"] = min_write_time
        return self.request(data)

    def delegators(
        self, account: str, threshold: int = 0, count: int = 0, start: str = ""
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#delegators"
        data: dict[str, Any] = {}
        data["action"] = "delegators"
        data["account"] = account
        if threshold:
            data["threshold"] = threshold
        if count:
            data["count"] = count
        if start:
            data["start"] = start
        return self.request(data)

    def delegators_count(self, account: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#delegators_count"
        data: dict[str, Any] = {}
        data["action"] = "delegators_count"
        data["account"] = account
        return self.request(data)

    def deterministic_key(self, seed: str, index: int) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#deterministic_key"
        data: dict[str, Any] = {}
        data["action"] = "deterministic_key"
        data["seed"] = seed
        data["index"] = index
        return self.request(data)

    def election_statistics(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#election_statistics"
        data: dict[str, Any] = {}
        data["action"] = "election_statistics"
        return self.request(data)

    def epoch_upgrade(
        self, epoch: int, key: str, count: int = 0, threads: int = 0
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#epoch_upgrade"
        data: dict[str, Any] = {}
        data["action"] = "epoch_upgrade"
        data["epoch"] = epoch
        data["key"] = key
        if count:
            data["count"] = count
        if threads:
            data["threads"] = threads
        return self.request(data)

    def frontier_count(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#frontier_count"
        data: dict[str, Any] = {}
        data["action"] = "frontier_count"
        return self.request(data)

    def frontiers(self, account: str, count: int = 1) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#frontiers"
        data: dict[str, Any] = {}
        data["action"] = "frontiers"
        data["account"] = account
        data["count"] = count
        return self.request(data)

    def keepalive(self, address: str, port: int) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#keepalive"
        data: dict[str, Any] = {}
        data["action"] = "keepalive"
        data["address"] = address
        data["port"] = port
        return self.request(data)

    def key_create(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#key_create"
        data: dict[str, Any] = {}
        data["action"] = "key_create"
        return self.request(data)

    def key_expand(self, key: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#key_expand"
        data: dict[str, Any] = {}
        data["action"] = "key_expand"
        data["key"] = key
        return self.request(data)

    def ledger(
        self,
        account: str,
        count: int = 1,
        representative: bool = False,
        weight: bool = False,
        receivable: bool = False,
        modified_since: int = 0,
        sorting: bool = False,
        threshold: int = 0,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#ledger"
        data: dict[str, Any] = {}
        data["action"] = "ledger"
        data["account"] = account
        data["count"] = count
        if representative:
            data["representative"] = True
        if weight:
            data["weight"] = True
        if receivable:
            data["receivable"] = True
        if modified_since:
            data["modified_since"] = modified_since
        if sorting:
            data["sorting"] = True
        if threshold:
            data["threshold"] = threshold
        return self.request(data)

    def node_id(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#node_id"
        data: dict[str, Any] = {}
        data["action"] = "node_id"
        return self.request(data)

    def node_id_delete(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#node_id_delete"
        data: dict[str, Any] = {}
        data["action"] = "node_id_delete"
        return self.request(data)

    def peers(self, peer_details: bool = False) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#peers"
        data: dict[str, Any] = {}
        data["action"] = "peers"
        if peer_details:
            data["peer_details"] = True
        return self.request(data)

    def populate_backlog(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#populate_backlog"
        data: dict[str, Any] = {}
        data["action"] = "populate_backlog"
        return self.request(data)

    def process(
        self,
        block: str | dict[str, str],
        force: bool = False,
        subtype: str = "",
        json_block: bool = False,
        watch_work: bool = True,
        _async: bool = False,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#process"
        data: dict[str, Any] = {}
        data["action"] = "process"
        data["block"] = block
        if force:
            data["force"] = True
        if subtype:
            data["subtype"] = subtype
        if json_block:
            data["json_block"] = True
        if not watch_work:
            data["watch_work"] = False
        if _async:
            data["async"] = True
        return self.request(data)

    def receivable(
        self,
        account: str,
        count: int = 0,
        threshold: int = 0,
        source: bool = False,
        include_active: bool = False,
        min_version: bool = False,
        sorting: bool = False,
        include_only_confirmed: bool = True,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#receivable"
        data: dict[str, Any] = {}
        data["action"] = "receivable"
        data["account"] = account
        if count:
            data["count"] = count
        if threshold:
            data["threshold"] = threshold
        if source:
            data["source"] = True
        if include_active:
            data["include_active"] = True
        if min_version:
            data["min_version"] = True
        if sorting:
            data["sorting"] = True
        if not include_only_confirmed:
            data["include_only_confirmed"] = False
        return self.request(data)

    def receivable_exists(
        self,
        _hash: str,
        include_active: bool = False,
        include_only_confirmed: bool = True,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#receivable_exists"
        data: dict[str, Any] = {}
        data["action"] = "receivable_exists"
        data["hash"] = _hash
        if include_active:
            data["include_active"] = True
        if not include_only_confirmed:
            data["include_only_confirmed"] = False
        return self.request(data)

    def representatives(self, count: int = 1, sorting: bool = False) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#representatives"
        data: dict[str, Any] = {}
        data["action"] = "representatives"
        data["count"] = count
        if sorting:
            data["sorting"] = True
        return self.request(data)

    def representatives_online(
        self, weight: bool = False, accounts: list[str] | None = None
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#representatives_online"
        data: dict[str, Any] = {}
        data["action"] = "representatives_online"
        if weight:
            data["weight"] = True
        if accounts:
            data["accounts"] = accounts
        return self.request(data)

    def republish(
        self, _hash: str, count: int = 1, sources: int = 0, destinations: int = 0
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#republish"
        data: dict[str, Any] = {}
        data["action"] = "republish"
        data["hash"] = _hash
        if sources:
            data["sources"] = sources
            data["count"] = count
        if destinations:
            data["destinations"] = destinations
            data["count"] = count
        return self.request(data)

    def sign(
        self,
        key: str = "",
        wallet: str = "",
        account: str = "",
        block: str = "",
        _hash: str = "",
        json_block: bool = False,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#sign"
        data: dict[str, Any] = {}
        data["action"] = "sign"
        if key:
            data["key"] = key
        if wallet:
            data["wallet"] = wallet
        if account:
            data["account"] = account
        if isinstance(block, str):
            data["block"] = block
        else:
            data["block"] = json.dumps(block)
        if _hash:
            data["_hash"] = _hash
        if json_block:
            data["json_block"] = True
        return self.request(data)

    def stats(self, _type: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#stats"
        data: dict[str, Any] = {}
        data["action"] = "stats"
        data["type"] = _type
        return self.request(data)

    def stats_clear(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#stats_clear"
        data: dict[str, Any] = {}
        data["action"] = "stats_clear"
        return self.request(data)

    def stop(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#stop"
        data: dict[str, Any] = {}
        data["action"] = "stop"
        return self.request(data)

    def successors(
        self, block: str, count: int = 1, offset: int = 0, reverse: bool = False
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#successors"
        data: dict[str, Any] = {}
        data["action"] = "successors"
        data["block"] = block
        data["count"] = count
        if offset:
            data["offset"] = offset
        if reverse:
            data["reverse"] = True
        return self.request(data)

    def telemetry(self, raw: bool = False, address: int = 0, port: int = 7075) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#telemetry"
        data: dict[str, Any] = {}
        data["action"] = "telemetry"
        if raw:
            data["raw"] = True
        if address:
            data["address"] = address
            data["port"] = port
        return self.request(data)

    def validate_account_number(self, account: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#validate_account_number"
        data: dict[str, Any] = {}
        data["action"] = "validate_account_number"
        data["account"] = account
        return self.request(data)

    def version(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#version"
        data: dict[str, Any] = {}
        data["action"] = "version"
        return self.request(data)

    def unchecked(self, json_block: bool = False, count: int = 1) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#unchecked"
        data: dict[str, Any] = {}
        data["action"] = "unchecked"
        if json_block:
            data["json_block"] = True
        data["count"] = count
        return self.request(data)

    def unchecked_clear(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#unchecked_clear"
        data: dict[str, Any] = {}
        data["action"] = "unchecked_clear"
        return self.request(data)

    def uncheckedget(self, _hash: str, json_block: bool = False) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#uncheckedget"
        data: dict[str, Any] = {}
        data["action"] = "uncheckedget"
        data["hash"] = _hash
        if json_block:
            data["json_block"] = True
        return self.request(data)

    def unchecked_keys(self, key: str, count: int = 1, json_block: bool = False) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#unchecked_keys"
        data: dict[str, Any] = {}
        data["action"] = "unchecked_keys"
        data["key"] = key
        data["count"] = count
        if json_block:
            data["json_block"] = True
        return self.request(data)

    def unopened(self, account: str = "", count: int = 1, threshold: int = 0) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#unopened"
        data: dict[str, Any] = {}
        data["action"] = "unopened"
        if account:
            data["account"] = account
        if count:
            data["count"] = count
        if threshold:
            data["threshold"] = threshold
        return self.request(data)

    def uptime(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#uptime"
        data: dict[str, Any] = {}
        data["action"] = "uptime"
        return self.request(data)

    def work_cancel(self, _hash: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#work_cancel"
        data: dict[str, Any] = {}
        data["action"] = "work_cancel"
        data["hash"] = _hash
        return self.request(data)

    def work_generate(
        self,
        _hash: str,
        use_peers: bool = False,
        difficulty: str = "",
        multiplier: int = 0,
        account: str = "",
        version: str = "work_1",
        block: str = "",
        json_block: bool = False,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#work_generate"
        data: dict[str, Any] = {}
        data["action"] = "work_generate"
        data["hash"] = _hash
        if use_peers:
            data["use_peers"] = True
        if multiplier:
            data["multiplier"] = multiplier
        elif difficulty:
            data["difficulty"] = difficulty
        if account:
            data["account"] = account
        if version in ["work_1"]:
            data["version"] = version
        if block:
            data["block"] = block
            if json_block:
                data["json_block"] = json_block
        return self.request(data)

    def work_peer_add(self, address: str, port: int) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#work_peer_add"
        data: dict[str, Any] = {}
        data["action"] = "work_peer_add"
        data["address"] = address
        data["port"] = port
        return self.request(data)

    def work_peers(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#work_peers"
        data: dict[str, Any] = {}
        data["action"] = "work_peers"
        return self.request(data)

    def work_peers_clear(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#work_peers_clear"
        data: dict[str, Any] = {}
        data["action"] = "work_peers_clear"
        return self.request(data)

    def work_validate(
        self,
        work: str,
        _hash: str,
        difficulty: str = "",
        multiplier: int = 0,
        version: str = "work_1",
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#work_validate"
        data: dict[str, Any] = {}
        data["action"] = "work_validate"
        data["work"] = work
        data["hash"] = _hash
        if multiplier:
            data["multiplier"] = multiplier
        elif difficulty:
            data["difficulty"] = difficulty
        if version in ["work_1"]:
            data["version"] = version
        return self.request(data)

    def account_create(self, wallet: str, index: int = 0, work: bool = True) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_create"
        data: dict[str, Any] = {}
        data["action"] = "account_create"
        data["wallet"] = wallet
        if index:
            data["index"] = index
        if not work:
            data["work"] = False
        return self.request(data)

    def account_list(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_list"
        data: dict[str, Any] = {}
        data["action"] = "account_list"
        data["wallet"] = wallet
        return self.request(data)

    def account_move(self, wallet: str, source: str, accounts: list[str]) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_move"
        data: dict[str, Any] = {}
        data["action"] = "account_move"
        data["wallet"] = wallet
        data["source"] = source
        data["accounts"] = accounts
        return self.request(data)

    def account_remove(self, wallet: str, account: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_remove"
        data: dict[str, Any] = {}
        data["action"] = "account_remove"
        data["wallet"] = wallet
        data["account"] = account
        return self.request(data)

    def account_representative_set(
        self, wallet: str, account: str, representative: str, work: str = ""
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#account_representative_set"
        data: dict[str, Any] = {}
        data["action"] = "account_representative_set"
        data["wallet"] = wallet
        data["account"] = account
        data["representative"] = representative
        if work:
            data["work"] = work
        return self.request(data)

    def accounts_create(self, wallet: str, count: int = 1, work: bool = True) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#accounts_create"
        data: dict[str, Any] = {}
        data["action"] = "accounts_create"
        data["wallet"] = wallet
        data["count"] = count
        if not work:
            data["work"] = False
        return self.request(data)

    def password_change(self, wallet: str, password: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#password_change"
        data: dict[str, Any] = {}
        data["action"] = "password_change"
        data["wallet"] = wallet
        data["password"] = password
        return self.request(data)

    def password_enter(self, wallet: str, password: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#password_enter"
        data: dict[str, Any] = {}
        data["action"] = "password_enter"
        data["wallet"] = wallet
        data["password"] = password
        return self.request(data)

    def password_valid(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#password_valid"
        data: dict[str, Any] = {}
        data["action"] = "password_valid"
        data["wallet"] = wallet
        return self.request(data)

    def receive(self, wallet: str, account: str, block: str, work: str = "") -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#receive"
        data: dict[str, Any] = {}
        data["action"] = "receive"
        data["wallet"] = wallet
        data["account"] = account
        data["block"] = block
        if work:
            data["work"] = work
        return self.request(data)

    def receive_minimum(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#receive_minimum"
        data: dict[str, Any] = {}
        data["action"] = "receive_minimum"
        return self.request(data)

    def receive_minimum_set(self, amount: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#receive_minimum_set"
        data: dict[str, Any] = {}
        data["action"] = "receive_minimum_set"
        data["amount"] = amount
        return self.request(data)

    def search_receivable(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#search_receivable"
        data: dict[str, Any] = {}
        data["action"] = "search_receivable"
        data["wallet"] = wallet
        return self.request(data)

    def search_receivable_all(self) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#search_receivable_all"
        data: dict[str, Any] = {}
        data["action"] = "search_receivable_all"
        return self.request(data)

    def send(
        self,
        wallet: str,
        source: str,
        destination: str,
        amount: str,
        _id: str = "",
        work: str = "",
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#send"
        data: dict[str, Any] = {}
        data["action"] = "send"
        data["wallet"] = wallet
        data["source"] = source
        data["destination"] = destination
        data["amount"] = amount
        if _id:
            data["id"] = _id
        if work:
            data["work"] = work
        return self.request(data)

    def wallet_add(self, wallet: str, key: str, work: bool = False) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_add"
        data: dict[str, Any] = {}
        data["action"] = "wallet_add"
        data["wallet"] = wallet
        data["key"] = key
        if work:
            data["work"] = True
        return self.request(data)

    def wallet_add_watch(self, wallet: str, accounts: list[str]) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_add_watch"
        data: dict[str, Any] = {}
        data["action"] = "wallet_add_watch"
        data["wallet"] = wallet
        data["accounts"] = accounts
        return self.request(data)

    def wallet_balances(self, wallet: str, threshold: int = 0) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_balances"
        data: dict[str, Any] = {}
        data["action"] = "wallet_balances"
        data["wallet"] = wallet
        if threshold:
            data["threshold"] = threshold
        return self.request(data)

    def wallet_change_seed(self, wallet: str, seed: str, count: int = 0) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_change_seed"
        data: dict[str, Any] = {}
        data["action"] = "wallet_change_seed"
        data["wallet"] = wallet
        data["seed"] = seed
        if count:
            data["count"] = count
        return self.request(data)

    def wallet_contains(self, wallet: str, account: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_contains"
        data: dict[str, Any] = {}
        data["action"] = "wallet_contains"
        data["wallet"] = wallet
        data["account"] = account
        return self.request(data)

    def wallet_create(self, seed: str = "") -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_create"
        data: dict[str, Any] = {}
        data["action"] = "wallet_create"
        if seed:
            data["seed"] = seed
        return self.request(data)

    def wallet_destroy(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_destroy"
        data: dict[str, Any] = {}
        data["action"] = "wallet_destroy"
        data["wallet"] = wallet
        return self.request(data)

    def wallet_export(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_export"
        data: dict[str, Any] = {}
        data["action"] = "wallet_export"
        data["wallet"] = wallet
        return self.request(data)

    def wallet_frontiers(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_frontiers"
        data: dict[str, Any] = {}
        data["action"] = "wallet_frontiers"
        data["wallet"] = wallet
        return self.request(data)

    def wallet_history(self, wallet: str, modified_since: int = 0) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_history"
        data: dict[str, Any] = {}
        data["action"] = "wallet_history"
        data["wallet"] = wallet
        if modified_since:
            data["modified_since"] = modified_since
        return self.request(data)

    def wallet_info(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_info"
        data: dict[str, Any] = {}
        data["action"] = "wallet_info"
        data["wallet"] = wallet
        return self.request(data)

    def wallet_ledger(
        self,
        wallet: str,
        representative: bool = False,
        weight: bool = False,
        receivable: bool = False,
        modified_since: str = "",
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_ledger"
        data: dict[str, Any] = {}
        data["action"] = "wallet_ledger"
        data["wallet"] = wallet
        if representative:
            data["representative"] = True
        if weight:
            data["weight"] = True
        if receivable:
            data["receivable"] = True
        if modified_since:
            data["modified_since"] = modified_since
        return self.request(data)

    def wallet_lock(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_lock"
        data: dict[str, Any] = {}
        data["action"] = "wallet_lock"
        data["wallet"] = wallet
        return self.request(data)

    def wallet_locked(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_locked"
        data: dict[str, Any] = {}
        data["action"] = "wallet_locked"
        data["wallet"] = wallet
        return self.request(data)

    def wallet_receivable(
        self,
        wallet: str,
        count: int = 1,
        threshold: int = 0,
        source: bool = False,
        include_active: bool = False,
        min_version: bool = False,
        include_only_confirmed: bool = True,
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_receivable"
        data: dict[str, Any] = {}
        data["action"] = "wallet_receivable"
        data["wallet"] = wallet
        data["count"] = count
        if threshold:
            data["threshold"] = threshold
        if source:
            data["source"] = True
        if include_active:
            data["include_active"] = True
        if min_version:
            data["min_version"] = True
        if not include_only_confirmed:
            data["include_only_confirmed"] = False
        return self.request(data)

    def wallet_representative(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_representative"
        data: dict[str, Any] = {}
        data["action"] = "wallet_representative"
        data["wallet"] = wallet
        return self.request(data)

    def wallet_representative_set(
        self, wallet: str, representative: str, update_existing_accounts: bool = False
    ) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_representative_set"
        data: dict[str, Any] = {}
        data["action"] = "wallet_representative_set"
        data["wallet"] = wallet
        data["representative"] = representative
        if update_existing_accounts:
            data["update_existing_accounts"] = True
        return self.request(data)

    def wallet_republish(self, wallet: str, count: int = 1) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_republish"
        data: dict[str, Any] = {}
        data["action"] = "wallet_republish"
        data["wallet"] = wallet
        data["count"] = count
        return self.request(data)

    def wallet_work_get(self, wallet: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#wallet_work_get"
        data: dict[str, Any] = {}
        data["action"] = "wallet_workget"
        data["wallet"] = wallet
        return self.request(data)

    def work_get(self, wallet: str, account: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#workget"
        data: dict[str, Any] = {}
        data["action"] = "workget"
        data["wallet"] = wallet
        data["account"] = account
        return self.request(data)

    def work_set(self, wallet: str, account: str, work: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#work_set"
        data: dict[str, Any] = {}
        data["action"] = "work_set"
        data["wallet"] = wallet
        data["account"] = account
        data["work"] = work
        return self.request(data)

    def nano_to_raw(self, amount: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#nano_to_raw"
        data: dict[str, Any] = {}
        data["action"] = "nano_to_raw"
        data["amount"] = amount
        return self.request(data)

    def raw_to_nano(self, amount: str) -> Any:
        "https://docs.nano.org/commands/rpc-protocol/#raw_to_nano"
        data: dict[str, Any] = {}
        data["action"] = "raw_to_nano"
        data["amount"] = amount
        return self.request(data)


class HTTP(RPC):
    """HTTP RPC class

    :arg url: URL of the nano node
    """

    def __init__(self, url: str = "http://localhost:7076"):
        if not ENABLE_HTTP:
            raise ModuleNotFoundError("This class requires 'requests'.")
        self.url = url
        self.api = requests.session()

    def request(self, data: dict[str, Any]) -> Any:
        """Overridden from base class

        :arg data: dict like object
        :return: JSON reponse as dict
        """
        r = self.api.post(self.url, json=data)
        r.raise_for_status()
        return r.json()


class WS(RPC):
    """WS RPC class

    :arg url: URL of the nano node
    """

    def __init__(self, url: str = "ws://localhost:7078"):
        if not ENABLE_WS:
            raise ModuleNotFoundError("This class requires 'websocket-client'.")
        self.api = websocket.create_connection(url)

    def __del__(self) -> None:
        if ENABLE_WS:
            self.api.close()

    def request(self, data: dict[str, Any]) -> Any:
        """Overridden from base class

        :arg data: dict like object
        :return: JSON reponse as dict
        """
        self.api.send(json.dumps(data))
        return json.loads(self.api.recv())



"""
nanopy
######
"""

import base64
import binascii
import dataclasses
import decimal
import hashlib
import hmac
import json
import os
from typing import Optional
from . import ext  # type: ignore

decimal.setcontext(decimal.BasicContext)
decimal.getcontext().traps[decimal.Inexact] = True
decimal.getcontext().traps[decimal.Subnormal] = True
decimal.getcontext().prec = 40


def deterministic_key(seed: str, i: int = 0) -> str:
    """Derive deterministic private key from seed based on index i

    :arg seed: 64 hex char seed
    :arg i: index number, 0 to 2^32 - 1
    :return: 64 hex char private key
    """
    assert len(bytes.fromhex(seed)) == 32
    assert 0 <= i <= 1 << 32
    return hashlib.blake2b(
        bytes.fromhex(seed) + i.to_bytes(4, byteorder="big"), digest_size=32
    ).hexdigest()


try:
    import mnemonic

    def generate_mnemonic(strength: int = 256, language: str = "english") -> str:
        """Generate a BIP39 type mnemonic. Requires `mnemonic <https://pypi.org/project/mnemonic>`_

        :arg strength: choose from 128, 160, 192, 224, 256
        :arg language: one of the installed word list languages
        :return: word list
        """
        m = mnemonic.Mnemonic(language)
        return m.generate(strength=strength)

    def mnemonic_key(
        words: str, i: int = 0, passphrase: str = "", language: str = "english"
    ) -> str:
        """Derive deterministic private key from mnemonic based on index i.
           Requires `mnemonic <https://pypi.org/project/mnemonic>`_

        :arg words: word list
        :arg i: account index
        :arg passphrase: passphrase to generate seed
        :arg language: word list language
        :return: 64 hex char private key
        """
        m = mnemonic.Mnemonic(language)
        assert m.check(words)
        key = b"ed25519 seed"
        msg = m.to_seed(words, passphrase)
        h = hmac.new(key, msg, hashlib.sha512).digest()
        sk, key = h[:32], h[32:]
        for j in [44, 165, i]:
            j = j | 0x80000000
            msg = b"\x00" + sk + j.to_bytes(4, byteorder="big")
            h = hmac.new(key, msg, hashlib.sha512).digest()
            sk, key = h[:32], h[32:]
        return sk.hex()

except ModuleNotFoundError:  # pragma: no cover
    pass  # pragma: no cover


@dataclasses.dataclass
class Network:
    """Network

    :arg name: name of the network
    :arg prefix: prefix for accounts in the network
    :arg difficulty: base difficulty
    :arg send_difficulty: difficulty for send/change blocks
    :arg receive_difficulty: difficulty for receive/open blocks
    :arg exp: exponent to convert between raw and base currency unit
    :arg rpc_url: default RPC url for the network
    :arg std_unit: symbol or label for the default currency unit
    """

    name: str = "nano"
    prefix: str = "nano_"
    difficulty: str = "ffffffc000000000"
    send_difficulty: str = "fffffff800000000"
    receive_difficulty: str = "fffffe0000000000"
    exp: int = 30
    rpc_url: str = "http://localhost:7076"
    std_unit: str = "Ӿ"

    _d: type["decimal.Decimal"] = decimal.Decimal
    _b32std: bytes = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ234567"
    _b32nano: bytes = b"13456789abcdefghijkmnopqrstuwxyz"
    _nano2std: bytes = bytes.maketrans(_b32nano, _b32std)
    _std2nano: bytes = bytes.maketrans(_b32std, _b32nano)

    def from_multiplier(self, multiplier: float) -> str:
        """Get difficulty from multiplier

        :arg multiplier: positive number
        :return: 16 hex char difficulty
        """
        d = int((int(self.difficulty, 16) - (1 << 64)) / multiplier + (1 << 64))
        return f"{d:016x}"

    def to_multiplier(self, difficulty: str) -> float:
        """Get multiplier from difficulty

        :arg difficulty: 16 hex char difficulty
        :return: multiplier
        """
        if len(difficulty) != 16:
            raise ValueError("Difficulty should be 16 hex char")
        base_d = (1 << 64) - int(self.difficulty, 16)
        d = (1 << 64) - int(difficulty, 16)
        return base_d / d

    def from_pk(self, pk: str) -> str:
        """Get account address from public key

        :arg pk: 64 hex char public key
        """
        if len(pk) != 64:
            raise ValueError("Public key should be 64 hex char")
        p = bytes.fromhex(pk)
        checksum = hashlib.blake2b(p, digest_size=5).digest()
        addr = base64.b32encode(b"000" + p + checksum[::-1])
        addr = addr.translate(self._std2nano)[4:]
        return self.prefix + addr.decode()

    def to_pk(self, addr: str) -> str:
        """Get public key from account address

        :arg addr: account address
        """
        if len(addr) != len(self.prefix) + 60:
            raise ValueError(f"Invalid address: {addr}")
        if addr[: len(self.prefix)] != self.prefix:
            raise ValueError(f"Invalid address: {addr}")
        pc = base64.b32decode((b"1111" + addr[-60:].encode()).translate(self._nano2std))
        p, checksum = pc[3:-5], pc[:-6:-1]
        if hashlib.blake2b(p, digest_size=5).digest() != checksum:
            raise ValueError(f"Invalid address: {addr}")
        return p.hex()

    def from_raw(self, raw: int, exp: int = 0) -> str:
        """Divide raw by 10^exp

        :arg raw: raw amount
        :arg exp: positive number
        :return: raw divided by 10^exp
        """
        if exp <= 0:
            exp = self.exp
        nano = self._d(raw) * self._d(self._d(10) ** -exp)
        return f"{nano.quantize(self._d(self._d(10) ** -exp)):.{exp}f}"

    def to_raw(self, val: str, exp: int = 0) -> int:
        """Multiply val by 10^exp

        :arg val: val
        :arg exp: positive number
        :return: val multiplied by 10^exp
        """
        if exp <= 0:
            exp = self.exp
        return int((self._d(val) * self._d(self._d(10) ** exp)).quantize(self._d(1)))


NANO = Network()


class Account:
    """Account

    :arg network: network of this account
    """

    def __init__(self, network: "Network" = NANO, addr: str = "") -> None:
        self._frontier = "0" * 64
        self.network = network
        self._pk = self.network.to_pk(addr) if addr else ""
        self._raw_bal = 0
        self._rep = self
        self._sk = ""

    def __repr__(self) -> str:
        return self.addr

    def __bool__(self) -> bool:
        try:
            return self.addr != ""
        except ValueError:
            return False

    @property
    def addr(self) -> str:
        "Account address"
        return self.network.from_pk(self._pk)

    @addr.setter
    def addr(self, addr: str) -> None:
        self._pk = self.network.to_pk(addr)
        self._sk = ""

    @property
    def pk(self) -> str:
        "64 hex char account public key"
        return self._pk

    @pk.setter
    def pk(self, key: str) -> None:
        assert len(bytes.fromhex(key)) == 32
        self._pk = key
        self._sk = ""

    @property
    def sk(self) -> str:
        "64 hex char account secret/private key"
        return self._sk

    @sk.setter
    def sk(self, key: str) -> None:
        assert len(bytes.fromhex(key)) == 32
        self._pk = ext.publickey(bytes.fromhex(key)).hex()
        self._sk = key

    @property
    def bal(self) -> str:
        "Account balance"
        return self.network.from_raw(self.raw_bal)

    @bal.setter
    def bal(self, val: str) -> None:
        self.raw_bal = self.network.to_raw(val)

    @property
    def raw_bal(self) -> int:
        "Account raw balance"
        return self._raw_bal

    @raw_bal.setter
    def raw_bal(self, val: int) -> None:
        if val < 0:
            raise ValueError("Balance cannot be < 0")
        if val >= 1 << 128:
            raise ValueError("Balance cannot be >= 2^128")
        self._raw_bal = val

    @property
    def frontier(self) -> str:
        "64 hex char account frontier block hash"
        return self._frontier

    @frontier.setter
    def frontier(self, frontier: str) -> None:
        assert len(bytes.fromhex(frontier)) == 32
        self._frontier = frontier

    @property
    def rep(self) -> "Account":
        "Account representative"
        return self._rep

    @rep.setter
    def rep(self, rep: "Account") -> None:
        if not rep:
            raise ValueError("Representative is not initialised")
        self._rep = rep

    @property
    def state(self) -> tuple[str, int, "Account"]:
        "State of the account (frontier block digest, raw balance, representative)"
        return self.frontier, self.raw_bal, self.rep

    @state.setter
    def state(self, value: tuple[str, int, "Account"]) -> None:
        self.frontier = value[0]
        self.raw_bal = value[1]
        self.rep = value[2]

    def change_rep(self, rep: "Account", work: str = "") -> "StateBlock":
        """Construct a signed change StateBlock with work

        :arg rep: representative account
        :arg work: 16 hex char work for the block
        :return: a signed change StateBlock
        """
        b = StateBlock(self, rep, self.raw_bal, self.frontier, "0" * 64)
        self._sign(b)
        if work:
            assert len(bytes.fromhex(work)) == 8
            b.work = work
        else:
            b.work_generate(self.network.send_difficulty)
        self.rep = rep
        self.frontier = b.digest
        return b

    def receive(
        self, digest: str, raw_amt: int, rep: Optional["Account"] = None, work: str = ""
    ) -> "StateBlock":
        """Construct a signed receive StateBlock with work

        :arg digest: 64 hex char hash digest of the receive block
        :arg raw_amt: raw amount to receive
        :arg rep: representative account
        :arg work: 16 hex char work for the block
        :return: a signed receive StateBlock
        """
        assert len(bytes.fromhex(digest)) == 32
        if raw_amt <= 0:
            raise ValueError("Amount must be a positive integer")
        final_raw_bal = self.raw_bal + raw_amt
        if final_raw_bal >= 1 << 128:
            raise ValueError("raw balance after receive cannot be >= 2^128")
        brep = rep if rep else self.rep
        b = StateBlock(self, brep, final_raw_bal, self.frontier, digest)
        self._sign(b)
        if work:
            assert len(bytes.fromhex(work)) == 8
            b.work = work
        else:
            b.work_generate(self.network.receive_difficulty)
        if rep:
            self.rep = rep
        self.raw_bal = final_raw_bal
        self.frontier = b.digest
        return b

    def send(
        self,
        to: "Account",
        raw_amt: int,
        rep: Optional["Account"] = None,
        work: str = "",
    ) -> "StateBlock":
        """Construct a signed send StateBlock with work

        :arg to: Destination account
        :arg raw_amt: raw amount to send
        :arg rep: representative account
        :arg work: 16 hex char work for the block
        :return: a signed send StateBlock
        """
        if not isinstance(raw_amt, int) or raw_amt <= 0:
            raise ValueError("Amount must be a positive integer")
        final_raw_bal = self.raw_bal - raw_amt
        if final_raw_bal < 0:
            raise ValueError("raw balance after send cannot be < 0")
        brep = rep if rep else self.rep
        b = StateBlock(self, brep, final_raw_bal, self.frontier, to.pk)
        self._sign(b)
        if work:
            assert len(bytes.fromhex(work)) == 8
            b.work = work
        else:
            b.work_generate(self.network.send_difficulty)
        if rep:
            self.rep = rep
        self.raw_bal = final_raw_bal
        self.frontier = b.digest
        return b

    def _sign(self, b: "StateBlock") -> None:
        """Sign a block

        :arg b: state block to be signed
        """
        if not self._sk:
            raise NotImplementedError("This method needs private key")
        h = bytes.fromhex(b.digest)
        s = bytes.fromhex(self._sk)
        b.sig = str(ext.sign(s, h, os.urandom(32)).hex())


@dataclasses.dataclass
class StateBlock:
    """State block

    :arg acc: account of the block
    :arg rep: account representative
    :arg bal: account balance
    :arg prev: 64 hex char hash digest of the previous block
    :arg link: 64 hex char block link
    :arg sig: 128 hex char block signature
    :arg work: 16 hex char block work
    """

    acc: Account
    rep: Account
    bal: int
    prev: str
    link: str
    sig: str = ""
    work: str = ""

    @property
    def digest(self) -> str:
        "64 hex char hash digest of block"
        h = f"{'0' * 63}6{self.acc.pk}{self.prev}{self.rep.pk}{self.bal:032x}{self.link}"
        return hashlib.blake2b(bytes.fromhex(h), digest_size=32).hexdigest()

    @property
    def json(self) -> str:
        "block as JSON string"
        d = {
            "type": "state",
            "account": self.acc.addr,
            "previous": self.prev,
            "representative": self.rep.addr,
            "balance": self.bal,
            "link": self.link,
            "work": self.work,
            "signature": self.sig,
        }
        return json.dumps(d)

    def verify_signature(self) -> bool:
        """Verify signature for block

        :return: True if valid, False otherwise
        """
        s = bytes.fromhex(self.sig)
        p = bytes.fromhex(self.acc.pk)
        h = bytes.fromhex(self.digest)
        return bool(ext.verify_signature(s, p, h))

    def work_generate(self, difficulty: str) -> None:
        """Compute work

        :arg difficulty: 16 hex char difficulty
        """
        assert len(bytes.fromhex(difficulty)) == 8
        w = ext.work_generate(bytes.fromhex(self.prev), int(difficulty, 16))
        self.work = f"{w:016x}"

    def work_validate(self, difficulty: str) -> bool:
        """Check whether block has a valid work.

        :arg difficulty: 16 hex char difficulty
        :arg multiplier: positive number, overrides difficulty
        """
        assert len(bytes.fromhex(difficulty)) == 8
        h = bytes.fromhex(self.prev)
        return bool(ext.work_validate(int(self.work, 16), h, int(difficulty, 16)))



import os
import platform
import setuptools
import sysconfig

k, _, _, _, m, _ = platform.uname()
cc = os.getenv("CC", sysconfig.get_config_var("CC"))
print(k, m, cc)
BLAKE2B_DIR = "src/nanopy/blake2b/"
ED25519_DIR = "src/nanopy/ed25519-donna"
ED25519_SRC = ED25519_DIR + "/ed25519.c"
ED25519_IMPL = None
if m.lower() in ["x86_64", "amd64"]:
    BLAKE2B_DIR += "sse"
    ED25519_IMPL = "ED25519_SSE2"
elif m.lower().startswith("arm64") or m.lower().startswith("aarch64"):
    BLAKE2B_DIR += "neon"
else:
    BLAKE2B_DIR += "ref"
BLAKE2B_SRC = BLAKE2B_DIR + "/blake2b.c"

e_args = {
    "name": "nanopy.ext",
    "sources": ["src/nanopy/ext.c", BLAKE2B_SRC, ED25519_SRC],
    "include_dirs": [BLAKE2B_DIR, ED25519_DIR],
    "define_macros": [],
    "extra_compile_args": [],
    "extra_link_args": [],
    "libraries": [],
}

if k == "Windows":
    e_args["extra_compile_args"] += ["/arch:SSE2", "/arch:AVX", "/arch:AVX2"]
else:
    e_args["extra_compile_args"] += ["-O3", "-flto", "-march=native"]
    e_args["extra_link_args"] += ["-O3", "-flto", "-march=native", "-s"]

if ED25519_IMPL:
    e_args["define_macros"] += [(ED25519_IMPL, None)]

if os.environ.get("USE_GPU"):
    if k == "Darwin":
        e_args["define_macros"] += [("HAVE_OPENCL_OPENCL_H", None)]
        e_args["extra_link_args"] += ["-framework", "OpenCL"]
    else:
        e_args["define_macros"] += [("HAVE_CL_CL_H", None)]
        e_args["libraries"] += ["OpenCL"]
elif k == "Windows":
    e_args["extra_compile_args"] += ["/openmp"]
else:
    e_args["extra_compile_args"] += ["-fopenmp"]
    e_args["extra_link_args"] += ["-fopenmp"]

print(e_args)
setuptools.setup(ext_modules=[setuptools.Extension(**e_args)])
